# Copyright 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .utils.import_utils import (
    is_veomni_patch_available,
)
from .utils.logging import get_logger


logger = get_logger(__name__)


def _safe_apply_patches():
    if is_veomni_patch_available():
        from veomni_patch import apply_patch

        apply_patch()
        logger.info_rank0("✅ veomni_patch is available")
    else:
        logger.info_rank0("❌ veomni_patch is not available")


_safe_apply_patches()

__version__ = "v0.1.4"
